import urllib

import pytest
from django.urls import reverse

from lingo.epayment.models import PaymentBackend, Transaction

pytestmark = pytest.mark.django_db


def test_payment_disabled_demo(app, settings):
    settings.DEBUG = False
    app.get(reverse('lingo-epayment-demo'), status=404)


def test_payment_redirect(app, settings):
    settings.DEBUG = True
    PaymentBackend.objects.create(label='Test', service='dummy')
    resp = app.get(reverse('lingo-epayment-demo'))
    assert resp.location.startswith('https://dummy-payment.entrouvert.com/?')
    assert Transaction.objects.all().count() == 1
    transaction = Transaction.objects.all().first()
    assert transaction.status == 0
    assert transaction.amount == 20


def test_payment_redirect_select_backend(app, settings):
    settings.DEBUG = True
    PaymentBackend.objects.create(label='Test', service='dummy')
    resp = app.get(reverse('lingo-epayment-demo') + '?backend=test')
    assert resp.location.startswith('https://dummy-payment.entrouvert.com/?')
    assert Transaction.objects.all().count() == 1
    transaction = Transaction.objects.all().first()
    assert transaction.status == 0
    assert transaction.amount == 20


def test_payment_redirect_by_form(app, settings):
    settings.DEBUG = True
    PaymentBackend.objects.create(
        label='Test',
        service='ogone',
        service_options={'environment': 'TEST', 'pspid': 'xxx', 'sha_in': 'xxx', 'sha_out': 'xxx'},
    )
    resp = app.get(reverse('lingo-epayment-demo'))
    assert resp.pyquery('input[name="AMOUNT"]').attr.value == '2000'
    assert Transaction.objects.all().count() == 1
    transaction = Transaction.objects.all().first()
    assert transaction.status == 0
    assert transaction.amount == 20


def test_payment_return(app, settings):
    settings.DEBUG = True
    PaymentBackend.objects.create(label='Test', service='dummy')
    app.get(reverse('lingo-epayment-demo'))  # init transaction
    transaction = Transaction.objects.all().first()
    resp = app.get(
        reverse('lingo-epayment-explicit-return', kwargs={'transaction_id': transaction.id})
        + f'?transaction_id={transaction.order_id}&origin=origin&ok=1&signed=1'
    ).follow()
    status_url = reverse('lingo-epayment-processing-status', kwargs={'transaction_id': transaction.id})
    assert status_url in resp
    assert resp.pyquery('#continue a').attr.href == '/'
    assert app.get(status_url).json == {'status': 3, 'running': False, 'paid': True}


def test_payment_return_next_url(app, settings):
    settings.DEBUG = True
    PaymentBackend.objects.create(label='Test', service='dummy')
    app.get(reverse('lingo-epayment-demo') + '?next_url=/next/')  # init transaction
    transaction = Transaction.objects.all().first()
    resp = app.get(
        reverse('lingo-epayment-explicit-return', kwargs={'transaction_id': transaction.id})
        + f'?transaction_id={transaction.order_id}&origin=origin&ok=1&signed=1'
    ).follow()
    status_url = reverse('lingo-epayment-processing-status', kwargs={'transaction_id': transaction.id})
    assert status_url in resp
    assert resp.pyquery('#continue a').attr.href == '/next/'
    assert app.get(status_url).json == {'status': 3, 'running': False, 'paid': True}


def test_payment_callback(app, settings):
    settings.DEBUG = True
    PaymentBackend.objects.create(label='Test', service='dummy')
    app.get(reverse('lingo-epayment-demo'))  # init transaction
    transaction = Transaction.objects.all().first()
    app.get(
        reverse('lingo-epayment-explicit-callback', kwargs={'transaction_id': transaction.id})
        + f'?transaction_id={transaction.order_id}&origin=origin&ok=1&signed=1',
        status=200,
    )
    transaction.refresh_from_db()
    assert transaction.status == 3


def test_payment_auto_return(app, settings):
    settings.DEBUG = True
    app.get(reverse('lingo-epayment-auto-return'), status=400)

    backend = PaymentBackend.objects.create(
        label='Test',
        service='ogone',
        service_options={
            'environment': 'TEST',
            'pspid': 'xxx',
            'sha_in': 'xxx',
            'sha_out': 'xxx',
            'encoding': 'utf-8',
        },
    )
    app.get(reverse('lingo-epayment-demo'))  # init transaction
    transaction = Transaction.objects.all().first()

    ogone_backend = backend.eopayment
    data = {
        'orderid': transaction.order_id,
        'status': '9',
        'payid': '3011229363',
        'ncerror': '0',
    }
    data['shasign'] = ogone_backend.backend.sha_sign_out(data, encoding='utf-8')
    resp = app.get(reverse('lingo-epayment-auto-return') + '?' + urllib.parse.urlencode(data))
    assert resp.location == reverse('lingo-epayment-processing', kwargs={'transaction_id': transaction.id})
    transaction.refresh_from_db()
    assert transaction.status == 3


def test_payment_auto_callback(app, settings):
    settings.DEBUG = True
    assert app.get(reverse('lingo-epayment-auto-callback')).json['err'] == 1

    backend = PaymentBackend.objects.create(
        label='Test',
        service='ogone',
        service_options={
            'environment': 'TEST',
            'pspid': 'xxx',
            'sha_in': 'xxx',
            'sha_out': 'xxx',
            'encoding': 'utf-8',
        },
    )
    app.get(reverse('lingo-epayment-demo'))  # init transaction
    transaction = Transaction.objects.all().first()

    ogone_backend = backend.eopayment
    data = {
        'orderid': transaction.order_id,
        'status': '9',
        'payid': '3011229363',
        'ncerror': '0',
    }
    data['shasign'] = ogone_backend.backend.sha_sign_out(data, encoding='utf-8')
    assert (
        app.get(reverse('lingo-epayment-auto-callback') + '?' + urllib.parse.urlencode(data)).json['err'] == 0
    )
    transaction.refresh_from_db()
    assert transaction.status == 3
